from __future__ import division
from random import random, randint
from math import *

class Board:
    def __init__(self):
        self.grid = []
        for i in range(20 * 20):
            self.grid.append([])

    def get_terrain_height(self, pos):
        return 0

    def get_terrain_type(self, pos):
        return 0

    def get_area_type(self, pos):
        return 0

    def get_objects(self, pos, r):
        cs = []
        px, py = pos
        x_, y_ = int((px + 1) * 10) - 1, int((py + 1) * 10) - 1
        for i in range(3):
            for j in range(3):
                x, y = x_ + i, y_ + j
                if x < 0 or y < 0 or x > 19 or y > 19: continue
                for c in self.grid[y * 20 + x]:
                    cx, cy = c.pos
                    q = (cx - px) ** 2 + (cy - py) ** 2
                    if q < (r + c.radius) ** 2:
                        cs.append(c)
        return cs

    def get_grid(self, o):
        x, y = o.pos
        x, y = int((x + 1) * 10), int((y + 1) * 10)
        try:
            return self.grid[y * 20 + x]
        except IndexError:
            print("invalid object at %d/%d", x, y)

    def remove_object(self, o):
        self.get_grid(o).remove(o)

    def add_object(self, o):
        x, y = o.pos
        x, y = int((x + 1) * 10), int((y + 1) * 10)
        self.get_grid(o).append(o)

class Object:
    maxhp = 10
    def __init__(self, pos):
        self.pos = pos
        self.timer = 0
        self.to = pos
        self.dir = 0
        self.layer = 0
        self.removed = False
        self.hp = self.maxhp

    def tick(self):
        pass

class Game:
    def __init__(self):
        global game
        game = self
        self.ticks = 0
        self.board = Board()
        self.objects = []
        self.actors = []
        self.enemies = []
        self.buildings = []
        self.radius = 0.87
        self.pos = 0, 0

        self.build = None

        self.build_orders = []
        self.home_orders = []

        self.nectar = 0
        self.max_upkeep = 0
        self.upkeep = 0
        self.cave = None

        self.collector_appears_timer = 30 * 10

        self.start = True

        self.enum = 0
        self.max_enum = 21
        self.eticker = 30 * 60 * 2
        self.espeed = 30 * 30
        self.initial_nectar = 20

    def make(self, type, pos):
        o = type(pos)
        self.objects.append(o)
        self.board.add_object(o)
        return o

    def make_actor(self, type, pos):
        o = self.make(type, pos)
        self.actors.append(o)
        o.dir = random() * 2 * pi
        return o

    def make_enemy(self, type, pos):
        o = self.make(type, pos)
        self.enemies.append(o)
        x, y = pos
        o.dir = atan2(-x, -y)
        return o

    def make_building(self, type, pos):
        o = self.make(type, pos)
        self.buildings.append(o)
        o.built = False
        return o

    def remove(self, o):
        o.removed = True
        self.board.remove_object(o)
        self.objects.remove(o)

    def move(self, o, pos):
        lx, ly = o.pos
        nx, ny = pos
        if lx != nx or ly != ny:
            self.board.remove_object(o)
            o.pos = pos
            self.board.add_object(o)
            o.dir = atan2(ny - ly, nx - lx)

    def can_build(self, pos):
        if self.start:
            if self.build != Cave: return False
        else:
            if self.build == Cave: return False
        if self.build == Remove:
            return True
        if self.build and self.nectar >= self.build.cost:
            build = self.build
            x, y = pos
            gx, gy = self.pos
            q = (x - gx) ** 2 + (y - gy) ** 2
            if q > self.radius ** 2: return False
            if not game.board.get_objects(pos, build.radius):
                return True
        return False

    def kill(self, o):
        self.remove(o)
        if o in game.buildings:
            game.buildings.remove(o)
        if o in game.actors:
            if o.home and o.home != game.cave and not o.home.removed:
                self.kill(o.home)
            game.actors.remove(o)
            if isinstance(o, Builder):
                if o.to and o in o.to.builders:
                    o.to.builders.remove(o)

    def click_remove(self, pos):
        for b in game.board.get_objects(pos, self.build.radius):
            if b.__class__ in [Flower]  or (isinstance(b, Building) and
                not b.__class__ in [Cave]):
                self.kill(b)
                break

    def click(self, pos):
        if self.can_build(pos):
            if self.build == Remove:
                self.click_remove(pos)
                self.build = None
                return
            b = self.make_building(self.build, pos)
            if b.__class__ == Cave:
                self.cave = b
                b.layer = -1
                self.start = False
            else:
                self.build_orders.append(b)
            b.builders = []
            self.build = None
            self.nectar -= b.cost

    def tick(self):
        self.ticks += 1

        if self.eticker > 0:
            if self.enum < self.max_enum: self.eticker -= 1
            if self.eticker == 0:
                def spawn():
                    a = random() * 2 * pi
                    x = cos(a) * 0.9
                    y = sin(a) * 0.9
                    e = self.make_enemy(Caterp, (x, y))
                    e.hp = e.maxhp = max(10, self.enum)
                    return e
                e = spawn()
                self.eticker = self.espeed
                if self.enum % 5 == 3:
                    e.straight = True
                    if self.espeed > 30 * 5:
                        self.espeed = max(30 * 5, self.espeed - 30 * 10)
                if self.enum % 7 == 6:
                    n = 2 + 2 ** (self.enum // 7)
                    for i in range(n):
                        spawn()
                    e = spawn()
                    e.straight = True
                self.enum += 1

        if self.start: return

        if self.cave.timer < self.cave.build_time:
            self.cave.timer += 1
            if self.cave.timer == self.cave.build_time:
                self.cave.built = True

                imp = self.make_actor(Builder, self.cave.pos)
                imp.home = self.cave
                imp.state = "idle"
                self.nectar = self.initial_nectar

        if self.collector_appears_timer > 0:
            self.collector_appears_timer -= 1
            if self.collector_appears_timer == 0:
                imp = self.make_actor(Collector, self.cave.pos)
                imp.home = self.cave
                imp.state = "idle"
                imp.round = 0

        r = randint(0, 3 + len(self.actors))
        i = 0
        self.upkeep = 0
        for o in self.actors:
            if o.removed: continue
            if i == r:
                if o.state == "idle":
                    x, y = o.pos
                    hx, hy = o.home.pos
                    if (hx - x) ** 2 + (hy - y) ** 2 > (
                        o.home.radius + o.radius + 0.1) ** 2:
                        o.state = "gohome"
                        o.to = o.home
                    else:
                        speed = 1 / 30 / 30
                        o.movement += 1
                        a = o.dir + (random() - 0.5) * 20 * pi / 180
                        dx, dy = cos(a), sin(a)
                        x, y = x + dx * speed, y + dy * speed
                        self.move(o, (x, y))
            o.tick()
            self.upkeep += o.upkeep
            i += 1

        for o in self.enemies:
            o.tick()

        self.max_upkeep = 0
        for o in self.buildings:
            if o.built:
                self.max_upkeep += o.upkeep

        for i in range(len(self.home_orders)):
            type, home = self.home_orders[i]
            if type.upkeep + self.upkeep <= self.max_upkeep:
                o = game.make_actor(type, self.cave.pos)
                o.state = "gohome"
                o.to = home
                o.home = home
                del self.home_orders[i]
                break

def select_building(type):
    game.build = type

class Imp(Object):
    upkeep = 1
    radius = 0.02

    def __init__(self, pos):
        Object.__init__(self, pos)
        self.movement = 0
        self.state = "idle"
        self.home = None

    def tick(self):
        if self.state == "idle":
            if self.home.removed:
                self.state = "gohome"
                self.to = game.cave
            else:
                self.idle()
        elif self.state == "rest":
            self.timer += 1
            if self.timer == self.rest_time:
                self.state = "idle"
        elif self.state in ["goto", "gohome"]:
            x, y = self.pos
            tx, ty = self.to.pos
            dx, dy = tx - x, ty - y
            d = (dx ** 2 + dy ** 2) ** 0.5
            if d > self.to.radius + self.radius:
                x += dx / d * self.speed
                y += dy / d * self.speed
                game.move(self, (x, y))
                self.movement += 1
            elif self.state == "goto":
                self.state = "work"
            else:
                if not self.home or self.home.removed:
                    if self.to == game.cave:
                        game.remove(self)
                        game.actors.remove(self)
                    else:
                        self.home = None
                        self.to = game.cave
                else:
                    self.state = "rest"
                    self.timer = 0
        else:
            self.work()

class Builder(Imp):
    rest_time = 30 * 3
    speed = 2 / 30 / 30

    def idle(self):
        found = None
        for i in range(4):
            for b in game.build_orders:
                if len(b.builders) == i:
                    found = b
                    break
            else:
                continue
            break

        if not found:
            for b in game.buildings:
                if b.hp < b.maxhp and len(b.builders) == 0:
                    found = b
                    break

        if found:
            self.to = found
            self.state = "goto"
            found.builders.append(self)

    def work(self):
        b = self.to
        if b.timer < b.build_time:
            b.timer += 1
            if b.removed:
                b.timer = b.build_time
            if b.timer == b.build_time:
                if not b.removed:
                    game.home_orders.append((b.inhabitant, b))
                game.build_orders.remove(b)
                b.built = True
                b.builders.remove(self)
                self.state = "gohome"
                self.to = self.home
        elif b.hp < b.maxhp:
            b.hp += 1
            if b.hp == b.maxhp:
                b.builders.remove(self)
                self.state = "gohome"
                self.to = self.home
        else:
            self.state = "gohome"
            self.to = self.home

class Planter(Imp):
    rest_time = 30 * 3
    speed = 1 / 30 / 30

    def idle(self):
        x, y = self.home.pos
        a = random() * 2 * pi
        d = self.radius + PlanterBush.radius + random() * 0.08
        x += cos(a) * d
        y += sin(a) * d
        if not game.board.get_objects((x, y), Flower.radius):
            self.to = game.make(Flower, (x, y))
            self.to.picked = False
            self.state = "goto"

    def work(self):
        b = self.to
        if b.timer < b.build_time and not b.removed:
            b.timer += 1
        else:
            self.state = "gohome"
            self.to = self.home

class Collector(Imp):
    rest_time = 30 * 10
    speed = 1 / 30 / 30

    def idle(self):
        found = None
        for b in game.buildings:
            if b.round < self.round and b.nectar and b.built:
                found = b
                break

        if found:
            self.to = found
            self.state = "goto"
        else:
            
            self.round += 1

    def work(self):
        b = self.to
        b.round = self.round
        game.nectar += b.nectar
        self.state = "gohome"
        self.to = self.home

class Elf(Imp):
    upkeep = 1
    rest_time = 30 * 1
    speed = 2 / 30 / 30
    radius = 0.04

    def idle(self):
        x, y = self.home.pos
        a = random() * 2 * pi
        d = self.radius + Tree.radius + random() * 0.2
        x += cos(a) * d
        y += sin(a) * d
        for o in game.board.get_objects((x, y), 0.1):
            if o.__class__ == Flower:
                if o.timer == o.build_time and not o.picked:
                    self.to = o
                    self.picked = True
                    self.state = "goto"

    def work(self):
        b = self.to
        if not b.removed:
            game.remove(b)
            game.nectar += 1
        self.state = "gohome"
        self.to = self.home

class DefenderElf(Imp):
    upkeep = 1
    rest_time = 30 * 1
    speed = 3 / 30 / 30
    radius = 0.04

    def idle(self):
        if self.hp < self.maxhp:
            self.hp += 1
        def distance(e):
            ex, ey = e.pos
            x, y = self.home.pos
            return (x - ex) ** 2 + (y - ey) ** 2

        for e in sorted(game.enemies, key = distance):
            self.to = e
            self.state = "goto"
            break
        else:
            self.state = "rest"
            self.timer = 0

    def work(self):
        if self.state == "fight":
            self.timer += 1
            if self.timer >= self.rest_time:
                self.state = "goto"
            return
        e = self.to
        if not e.removed:
            e.hp -= 1
            if e.hp == 0:
                game.remove(e)
                game.enemies.remove(e)
                self.maxhp += 5
                self.hp = self.maxhp
            else:
                self.state = "fight"
                self.timer = 0
                return
        self.state = "gohome"
        self.to = self.home

class HealerElf(Imp):
    upkeep = 1
    rest_time = 30 * 1
    speed = 5 / 30 / 30
    radius = 0.04

    def idle(self):

        def distance(e):
            ex, ey = e.pos
            x, y = self.home.pos
            return (x - ex) ** 2 + (y - ey) ** 2

        for e in sorted(game.actors, key = distance):
            if isinstance(e, Imp) and e.hp < e.maxhp and not e is self:
                self.to = e
                self.state = "goto"
                break
        else:
            self.state = "rest"
            self.timer = 0

    def work(self):
        a = self.to
        if not a.removed and a.hp < a.maxhp:
            a.hp = a.maxhp
            self.maxhp += 1
            self.hp = self.maxhp
        self.state = "gohome"
        self.to = self.home

class Building(Object):
    def __init__(self, pos):
        Object.__init__(self, pos)
        self.round = 0

class Cave(Building):
    build_time = 30 * 3
    inhabitant = Builder
    upkeep = 8
    cost = 0
    maxhp = 32
    nectar = 0
    radius = 0.05

class BuilderBush(Building):
    build_time = 30 * 8
    inhabitant = Builder
    upkeep = 0
    cost = 1
    maxhp = 2
    nectar = 1
    radius = 0.04

class PlanterBush(Building):
    build_time = 30 * 12
    inhabitant = Planter
    upkeep = 0
    cost = 2
    maxhp = 4
    nectar = 1
    radius = 0.04

class Flower(Object):
    build_time = 30 * 3
    radius = 0.02
    maxhp = 1

class Tree(Building):
    build_time = 30 * 32
    inhabitant = Elf
    upkeep = 8
    cost = 8
    maxhp = 16
    nectar = 1
    radius = 0.08

class DefenseTree(Building):
    build_time = 30 * 48
    inhabitant = DefenderElf
    upkeep = 0
    cost = 16
    maxhp = 16
    nectar = 1
    radius = 0.08

class HealerBush(Building):
    build_time = 30 * 48
    inhabitant = HealerElf
    upkeep = 0
    cost = 8
    maxhp = 4
    nectar = 1
    radius = 0.08

class Remove(Object):
    radius = 0.01

class Ant(Object):
    radius = 0.04

class Caterp(Object):
    radius = 0.03
    speed = 1.5 / 30 / 30
    rest_time = 30 * 1
    heal_time = 30 * 2

    def __init__(self, pos):
        Object.__init__(self, pos)
        self.movement = 0
        self.state = "circle"
        self.ccw = 1
        self.cr = 0.85
        self.wind = 360
        self.straight = False
        x, y = self.pos
        self.ca = atan2(y, x)
        self.ignore_collision = 0

    def tick(self):
        if self.state == "rest":
            self.timer += 1
            if self.timer >= self.heal_time:
                if self.hp < self.maxhp:
                    self.hp += 1
                self.state = "creep"
        elif self.state == "circle":
            if self.straight:
                r = 0.1 + random() * 0.5
                a = random() * 2 * pi
                x, y = cos(a) * r, sin(a) * r
                self.topos = x, y
                self.state = "creep"
            else:
                self.ca += 2 * pi / 64 * self.ccw
                x = cos(self.ca) * self.cr
                y = sin(self.ca) * self.cr
                self.topos = x, y
                self.state = "creep"
                self.wind -= 360 / 64
                if self.wind <= 0:
                    self.cr -= 0.1
                    self.wind = 360
        elif self.state == "creep":
            x, y = self.pos
            tx, ty = self.topos
            dx, dy = tx - x, ty - y
            d = (dx ** 2 + dy ** 2) ** 0.5
            if d > self.radius:
                dx *= self.speed / d
                dy *= self.speed / d
                x += dx
                y += dy
                game.move(self, (x, y))
                self.movement += 1

                if self.hp < self.maxhp:
                    self.state = "rest"
                    self.timer = 0

                if self.ignore_collision > 0:
                    self.ignore_collision -= 1

                for o in game.board.get_objects((x, y), self.radius):
                    if isinstance(o, (Building, Flower)):
                        self.to = o
                        self.state = "fight"
                        self.timer = 0
                    elif isinstance(o, Imp):
                        self.to = o
                        self.state = "fight"
                        self.timer = 0
                    elif isinstance(o, (Caterp)):
                        if not o is self and not self.straight:
                            if self.ignore_collision == 0:
                                self.ccw = -self.ccw
                                game.move(self, (x - dx, y - dy))
                                self.state = "circle"
                                self.ignore_collision = 30
            else:
                self.state = "circle"
        elif self.state == "goto":
            x, y = self.pos
            tx, ty = self.to.pos
            dx, dy = tx - x, ty - y
            d = (dx ** 2 + dy ** 2) ** 0.5
            if d > self.radius + self.to.radius:
                x += dx / d * self.speed
                y += dy / d * self.speed
                game.move(self, (x, y))
                self.movement += 1
            else:
                self.state = "fight"
                self.timer = 0
        elif self.state == "fight":
            if self.timer < self.rest_time:
                self.timer += 1
                return
            o = self.to
            if o.hp > 0 and not o.removed:
                o.hp -= 1
                if o.hp == 0:
                    game.kill(o)
                    self.to = None
                    self.state = "creep"
                    if isinstance(o, Imp):
                        self.maxhp += 1
                        self.hp += 1
                else:
                    self.state = "goto"
            else:
                self.state = "creep"

class Beetle(Object):
    radius = 0.06

class Centip(Object):
    radius = 0.09
