from __future__ import division
import os, re
from SDL import *
from SDL.ttf import *
from SDL.image import *
from OpenGL.GL import *
from ConfigParser import ConfigParser

datapath = "data"

class Image:
    cache = {}
    def from_surface(self, s, scale):
        def pot(x):
            y = 1
            while y < x: y *= 2
            return y

        self.w, self.h = s.w, s.h
        self.x, self.y = 0, 0
        self.texw = pot(self.w)
        self.texh = pot(self.h)
        self.ox, self.oy = 0, 0

        s2 = SDL_CreateRGBSurface(SDL_SWSURFACE,
            self.texw, self.texh, 32,
            0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000)

        SDL_FillRect(s2, SDL_Rect(0, 0, s2.w, s2.h), SDL_MapRGBA(s2.format, 0, 0, 0, 0))

        SDL_BlitSurface(s, SDL_Rect(0, 0, s.w, s.h), s2, SDL_Rect(0, 0, s.w, s.h))

        self.texture = glGenTextures(1)

        glBindTexture(GL_TEXTURE_2D, self.texture)

        glTexImage2D(GL_TEXTURE_2D, 0, 4, self.texw, self.texh, 0, GL_RGBA,
            GL_UNSIGNED_BYTE, s2.pixels.to_string())

        glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)

        self.scale = scale
        self.parent = None

    def sub(self, rect):
        img = Image()
        img.x, img.y, img.w, img.h = rect
        img.texture = self.texture
        img.texw, img.texh = self.texw, self.texh
        img.scale = self.scale
        img.ox, img.oy = self.ox, self.oy
        img.parent = self
        return img

    def draw(self, x, y):
        glBindTexture(GL_TEXTURE_2D, self.texture)
        x, y = x + self.ox * self.scale, y + self.oy * self.scale
        w, h = self.w * self.scale, self.h * self.scale
        tl, tt = self.x / self.texw, self.y / self.texh
        tr, tb = (self.x + self.w) / self.texw, (self.y + self.h) / self.texh
        self.v = [(x, y), (x, y - h), (x + w, y), (x + w, y - h)]
        self.t = [(tl, tt), (tl, tb), (tr, tt), (tr, tb)]
        glTexCoordPointerf(self.t)
        glVertexPointerf(self.v)
        glDrawArrays(GL_TRIANGLE_STRIP, 0, 4)
    
    @staticmethod
    def from_text(text, font, scale):
        surf = TTF_RenderText_Blended(font, text, SDL_Color(255, 255, 255))
        SDL_SetAlpha(surf, 0, 255)
        img = Image()
        img.from_surface(surf, scale)
        return img

    def __del__(self):
        if self.parent: return
        if self.texture:
            glDeleteTextures([self.texture])

    @staticmethod
    def from_ini(filename, scale):
        ini = ConfigParser()
        ini.read(os.path.join(datapath, filename))
        pics = {}
        for s in ini.sections():
            x = ini.getint(s, "x")
            y = ini.getint(s, "y")
            w = ini.getint(s, "w")
            h = ini.getint(s, "h")
            ox = ini.getint(s, "ox")
            oy = ini.getint(s, "oy")
            filename = ini.get(s, "texture")
            img = Image.from_file(filename, scale)
            fname = os.path.splitext(os.path.split(str(s))[1])[0]
            name, num = re.match(r"(\D+)(\d+)", fname).groups()
            num = int(num)
            sub = img.sub((x, y, w, h))
            sub.ox = -ox
            sub.oy = oy
            pics[(name, num)] = sub
        return pics

    @staticmethod
    def from_file(filename, scale):
        filename = os.path.join(datapath, filename)
        if not filename in Image.cache:
            img = Image()
            su = IMG_Load(filename)
            SDL_SetAlpha(su, 0, 255)
            img.from_surface(su, scale)
            Image.cache[filename] = img
            img.ox = -img.w / 2
            img.oy = img.h / 2
        return Image.cache[filename]

class Font:
    Center = 1
    count = 0
    def __init__(self, filename, size, scale):
        if Font.count == 0:
            TTF_Init()
        Font.count += 1
        self.font = TTF_OpenFont(os.path.join(datapath, filename), size)
        self.cache = {}
        self.scale = scale
        self.line_height = TTF_FontLineSkip(self.font) * scale

    def draw(self, text, x, y, flags = 0):
        if not text in self.cache:
            self.cache[text] = Image.from_text(text, self.font, self.scale)
        if flags & Font.Center:
            x -= self.cache[text].w / 2 * self.scale
            y += self.cache[text].h / 2 * self.scale
        x = int(x / self.scale) * self.scale
        y = int(y / self.scale) * self.scale
        self.cache[text].draw(x, y)

    def __del__(self):
        TTF_CloseFont(self.font)
        Font.count -= 1
        if Font.count == 0:
            TTF_Quit()
