#!/usr/bin/env python
from __future__ import division
from SDL import *
from game import *
from draw import *
from ui import *

def main():
    SDL_Init(SDL_INIT_EVERYTHING)
    SDL_EnableUNICODE(True)

    w, h = 640, 400
    screen = SDL_SetVideoMode(w, h, 32, SDL_OPENGL | SDL_RESIZABLE)

    ui = UI()
    game = Game()
    renderer = SimpleDraw()
    paused = False
    quit = False
    want_redraw = True
    f = 0
    FPS = 30
    mpos = 0, 0
    time_offset = SDL_GetTicks()
    while not quit:
        while 1:
            e = SDL_PollEventAndReturn()
            if not e: break
            if e.type == SDL_QUIT:
                quit = True
            elif e.type == SDL_VIDEORESIZE:
                w, h = e.w, e.h
                screen = SDL_SetVideoMode(w, h, 32, SDL_OPENGL | SDL_RESIZABLE)
                renderer.resize()
            elif e.type == SDL_KEYDOWN:
                if e.keysym.sym == SDLK_ESCAPE:
                    quit = True
                elif e.keysym.unicode == "p":
                    paused = not paused
            elif e.type == SDL_MOUSEBUTTONDOWN:
                pos = ((e.x - w / 2) / (w / 2), (h / 2 - e.y) / (w / 2) * 2)
                ui.click(pos) or game.click(pos)
            elif e.type == SDL_MOUSEMOTION:
                mpos = ((e.x - w / 2) / (w / 2), (h / 2 - e.y) / (w / 2) * 2)

        # Tick at constant FPS.
        t = SDL_GetTicks() - time_offset
        while f * 1000.0 / FPS <= t:
            if not paused:
                game.tick()
            f += 1
            want_redraw = True

        # Redraw.
        if want_redraw:
            want_redraw = False
            t = SDL_GetTicks() - time_offset
            renderer.draw_game(game)
            renderer.draw_ui(ui, game, mpos)
            SDL_GL_SwapBuffers()
            #SDL_Flip(screen)
        else:
            SDL_Delay(1)

    SDL_Quit()

if __name__ == "__main__":
    main()
